"""
Create TELBAMS given a set of BAM files.

TELBAMS are comprised of all the reads in a BAM file that include the
canonic TTAGGG telomere sequence

Author: jhrf
"""

import os
import textwrap
from argparse import SUPPRESS
import telomerecat.telbam as telbam

import parabam

from . import add_arg, exit_with_msg

# Here we define what constitutes a telomereic read.
# If both reads in a pair match the critea we store
# them to the telbam
def rule(reads, constants, master):
  tel_pats = constants["tel_pats"]
  telomere_status = False

  for read in iter(reads):
    for pattern in tel_pats:
      if read.seq is not None and pattern in read.seq:
        telomere_status = True
        break

  results = []
  if telomere_status:
    results = [("telbam", reads[0]), ("telbam", reads[1])]
  return results


# This whole package is essentially just a wrapper for a call to parabam subset
class Bam2Telbam(parabam.core.Interface):
  """Interface to interact with telomerecat programatically. This interface
  is also called by the `telomerecat` script"""

  def __init__(
    self,
    temp_dir=None,
    task_size=250000,
    total_procs=4,
    reader_n=2,
    verbose=False,
    announce=True,
    cmd_run=False,
  ):

    super(Bam2Telbam, self).__init__(
      instance_name="telomerecat bam2telbam",
      temp_dir=temp_dir,
      task_size=task_size,
      total_procs=total_procs,
      reader_n=reader_n,
      verbose=verbose,
      announce=announce,
      cmd_run=cmd_run,
    )

  def run_cmd(self):
    """Called from the master `telomerecat` script which handels the
    cmd line interface. Requires an argparse parser. Users should call
    the `run` function."""

    self.run(input_paths=self.cmd_args.input, outbam_dir=self.cmd_args.outbam_dir, reference=self.cmd_args.reference)

  def run(self, input_paths, outbam_dir=None, reference=None):
    """The main function for invoking the part of the
       program which creates a telbam from a bam

    Arguments:
      bams (list): The BAM files we wish to run telomerecat telbam on
      total_procs (int): The maximum numbers of task that will be run at one time
      task_size (int): The amount of reads that any one task will process concurrently
      verbose (int): Expects an int from 0 to 2.
               The level of output produced by telomerecat
      keep_in_temp (bool): Files will be kept in temp file after processing.
                 Useful for incorporation into pipelines
      announce (bool): Specify whether the program returns a welcome string."""

    self.__introduce__()

    final_output_paths = {}

    if outbam_dir is None:
      keep_in_temp = True
      outbam_dir = self.temp_dir
    else:
      outbam_dir = os.path.normpath(outbam_dir)
      if os.path.exists(outbam_dir):
        if not os.access(outbam_dir, os.W_OK | os.X_OK):
          exit_with_msg(f"Error: do not have right permission to write into outbam_dir path: '{outbam_dir}'")
      else:
        os.makedirs(outbam_dir)

    telbam_paths = telbam.process_alignments(outbam_dir, self.total_procs, self.temp_dir, input_paths, reference=reference, verbose=self.verbose)
    final_output_paths.update(telbam_paths)

    self.__goodbye__()
    return final_output_paths

  def get_parser(self):
    parser = self.default_parser()
    parser.description = textwrap.dedent(
      """\
    %s
    %s

       The bam2telbam command allows you to generate a TELBAM
       from a parent BAM file. A TELBAM is a file including all of the
       reads with at least 2 occurences of the telomeric hexamer.

       Once you have genereated a TELBAM you may then generate length
       estimates more quickly, when compared to running the `bam2length`
       command on a full BAM file. This is helpful if you intend to
       generate TL estimates more than once or if you require a
       collection of telomere reads.

       Given the following BAM file:

         example_bam_name.bam

       `telomerecat bam2telbam` will create the following TELBAM in the
       directory which it is run:

         example_bam_name_telbam.bam

       To find out how to generate a length estimate from a TELBAM
       type `telomerecat telbam2length` into your terminal

    %s
     """
      % (self.instance_name, self.header_line, self.header_line,)
    )

    for arg_name in ['input_bam', 'outbam_dir', 'reference']:
      add_arg[arg_name](parser)

    parser.add_argument("-s", help=SUPPRESS)
    parser.add_argument("-f", help=SUPPRESS)

    return parser


if __name__ == "__main__":
  print("Type telomerecat -h for help!")

# ....happily ever after.



"""
Genereate length estimates for given BAM files.

This simply strings together bam2telbam
and telbam2length packages

Author: jhrf
"""

import textwrap
from argparse import SUPPRESS
from telomerecat.core import TelomerecatInterface

# import args
from . import add_arg

class Bam2Length(TelomerecatInterface):
  def __init__(
    self,
    temp_dir=None,
    task_size=250000,
    total_procs=4,
    reader_n=1,
    verbose=False,
    announce=True,
    cmd_run=False,
  ):

    super(Bam2Length, self).__init__(
      instance_name="telomerecat bam2length",
      temp_dir=temp_dir,
      task_size=task_size,
      total_procs=total_procs,
      reader_n=reader_n,
      verbose=verbose,
      announce=announce,
      cmd_run=cmd_run,
    )

  def run_cmd(self):
    # TODO: existence of self.cmd_args.outbam_dir should be checked before proceeding
    self.run(
      input_paths=self.cmd_args.input,
      output_path=self.cmd_args.output,
      inserts_path=self.cmd_args.insert,
      outbam_dir=self.cmd_args.outbam_dir,
      correct_f2a=self.cmd_args.enable_correction,
      simulator_n=self.cmd_args.simulator_runs,
      seed_randomness=self.cmd_args.seed_randomness
    )

  def run(
    self,
    input_paths,
    output_path=None,
    outbam_dir=None,
    inserts_path=None,
    correct_f2a=False,
    simulator_n=10,
    seed_randomness=False
  ):

    # Import here to avoid infinite loop on import
    from telomerecat.bam2telbam import Bam2Telbam
    from telomerecat.telbam2length import Telbam2Length

    self.__introduce__()

    telbam_interface = Bam2Telbam(
      temp_dir=self.temp_dir,
      total_procs=self.total_procs,
      task_size=self.task_size,
      reader_n=self.reader_n,
      verbose=self.verbose,
      announce=False,
    )

    out_files = telbam_interface.run(input_paths=input_paths, outbam_dir=outbam_dir)

    length_paths = self.collapse_out_files(out_files)
    length_interface = Telbam2Length(
      temp_dir=self.temp_dir,
      total_procs=self.total_procs,
      task_size=self.task_size,
      reader_n=self.reader_n,
      announce=False,
      verbose=self.verbose,
    )

    length_interface.run(
      input_paths=length_paths,
      output_path=output_path,
      inserts_path=inserts_path,
      simulator_n=simulator_n,
      correct_f2a=correct_f2a,
      seed_randomness=seed_randomness
    )

    self.__goodbye__()
    self.interface_exit()

  def collapse_out_files(self, out_files):
    length_paths = []
    for path in out_files.keys():
      length_paths.append(out_files[path]["telbam"])
    return length_paths

  def get_parser(self):
    parser = self.default_parser()
    parser.description = textwrap.dedent(
      """\
    %s
    %s

      The bam2length command allows the user to genereate a telomere
      length estimate from a BAM file.

      The majority of the time taken running the bam2length script is
      spent collecting telomeric reads from the BAM file. By default
      this command will retain the TELBAM generted as part
      of the analysis.

      If you wish to generate TELBAMS seperately from length estimation
      you should use the bam2telbam command.

      Type `telomerecat bam2telbam` to find out more.

    %s
    """
      % (self.instance_name, self.header_line, self.header_line,)
    )

    for arg_name in ['input_bam', 'outbam_dir', 'output_csv', 'trim', 'nreads_for_task', 'reference']:
      add_arg[arg_name](parser)

    parser.add_argument("-s", help=SUPPRESS)
    parser.add_argument("-f", help=SUPPRESS)

    return parser


if __name__ == "__main__":
  print("Please do not run this script directly. Type telomerecat -h for more information.")


# ....happily ever after.



import sys

# oh 42, the answer to everything.
RANDOM_SEED = 42

def add_input_telbam_arg(parser):
  parser.add_argument(
    "input", metavar="TELBAM(S)", nargs="+", help="The TELBAM(s) that we wish to analyse"
  )


def add_input_bam_arg(parser):
  parser.add_argument(
    "input",
    metavar="BAM/CRAM(S)",
    nargs="+",
    help=("BAM/CRAM file(s) for which we wish to\n" "generate telomere length estimates"),
  )

def add_ref_arg(parser):
  parser.add_argument(
    "-r",
    "--reference",
    metavar="FASTA",
    type=str,
    default=None,
    help=("Reference genome for CRAM inputs"),
  )


def add_trim_arg(parser):
  parser.add_argument(
    "-t",
    "--trim",
    type=int,
    metavar="INT",
    default=0,
    help="Use only the amount of sequence specified by this  \n"
    "option (i.e if the value 90 is supplied\n"
    "then only the first 90 bases are\n"
    "considered) [Default: Whole read]",
  )


def add_outbam_dir_arg(parser):
  parser.add_argument(
    "--outbam_dir",
    metavar="DIR",
    type=str,
    default=None,
    help=(
      "Output folder for bams which have all reads with at least 2 occurences\n"
      "of the telomeric hexamer. [Default: None, telbams will be discarded]"
    ),
  )


def add_output_csv_arg(parser):
  parser.add_argument(
    "--output",
    metavar="PATH",
    type=str,
    default="./telomerecat_length.csv",
    help=(
      "Specify output path for length estimation CSV.\n"
      "[Default: './telomerecat_length.csv']"
    ),
  )


def add_nreads_for_each_task_arg(parser):
  parser.add_argument(
    "-s",
    type=int,
    metavar="INT",
    default=250000,
    help=("The amount of reads considered by each\n" "distributed task. [Default: 250000]"),
  )

add_arg = {
  'input_telbam': add_input_telbam_arg,
  'input_bam': add_input_bam_arg,
  'trim': add_trim_arg,
  'outbam_dir': add_outbam_dir_arg,
  'output_csv': add_output_csv_arg,
  'nreads_for_task': add_nreads_for_each_task_arg,
  'reference': add_ref_arg
}

def exit_with_msg(message):
  print(message, file=sys.stderr)
  sys.exit(1)



"""
This module defines a class that all other commands inherit from.

It's primary responsibility is to define the default parameters for argparse

Author: jhrf
"""

from argparse import SUPPRESS
import parabam
from abc import ABCMeta


class TelomerecatInterface(parabam.core.Interface, metaclass=ABCMeta):

  def __init__(
    self,
    instance_name,
    temp_dir=None,
    task_size=10000,
    total_procs=4,
    reader_n=2,
    verbose=False,
    announce=True,
    cmd_run=False,
  ):

    super(TelomerecatInterface, self).__init__(
      instance_name=instance_name,
      temp_dir=temp_dir,
      task_size=task_size,
      total_procs=total_procs,
      reader_n=reader_n,
      verbose=verbose,
      announce=announce,
      cmd_run=cmd_run,
    )

  def __output__(self, outstr, level=-1):
    if self.verbose and (self.verbose >= level or level == -1):
      print(outstr)

  def default_parser(self):
    parser = super(TelomerecatInterface, self).default_parser()

    parser.add_argument(
      "--insert",
      metavar="CSV",
      type=str,
      default=None,
      help=(
        "A file specifying the insert length mean and\n"
        "std for each input sample. If not present\n"
        "telomerecat will automatically estimate\n"
        "insert length of sample [Default: None]"
      ),
    )
    parser.add_argument(
      "-N",
      "--simulator_runs",
      metavar="INT",
      type=int,
      default=10,
      help=(
        "The amount of times to run the length simulator.\n"
        "A higher number better captures the uncertainty \n"
        "produced by the insert length\n"
        "distribution [Default 10]"
      ),
    )
    parser.add_argument(
      "-e",
      "--enable_correction",
      action="store_true",
      default=False,
      help="Correction will be applied to F2a values",
    )

    parser.add_argument(
      "--seed_randomness",
      action="store_true",
      default=False,
      # Use the seed (42) for all randomnesses in telomerecat to produce stable results.
      help=SUPPRESS  # this option is hidden from users.
    )
    parser.add_argument(
        '-w', '--prior_weight',
        metavar='INT', type=int, default=3,
        help=('The weight given to the prior expectation\n'
              'in F2a correction [Default 3]'))

    return parser



"""
Create length estimates given a set of TELBAMS.

Author: jhrf
"""

import textwrap
import time
import os
import re
import parabam

import numpy as np
import random
import pandas as pd

# from itertools import izip
from collections import namedtuple

from telomerecat.csv2length import Csv2Length
from telomerecat.core import TelomerecatInterface

# import args
from . import add_arg
from functools import partial

from . import RANDOM_SEED

class SimpleReadFactory(object):
  def __init__(self, vital_stats=None, trim_reads=0):
    self._SimpleRead = namedtuple(
      "SimpleRead", "seq qual" + " five_prime pattern mima_loci n_loci" + " avg_qual"
    )

    if vital_stats:
      self._read_len = vital_stats["read_len"]
      self._phred_offset = vital_stats["phred_offset"]
    else:
      self._read_len = 100
      self._phred_offset = 33

    self._trim_reads = trim_reads
    self._compliments = {"A": "T", "T": "A", "C": "G", "G": "C", "N": "N"}

    self.mima_logic = MismatchingLociLogic()

  def get_simple_read(self, read):
    seq, qual = self.__flip_and_compliment__(read)
    if self._trim_reads > 0:
      seq, qual = (seq[: self._trim_reads], qual[: self._trim_reads])

    (mima_loci, frameshift_loci), pattern = self.mima_logic.get_telo_mismatch(seq, qual)

    avg_qual, n_loci = self.__get_phred_score__(seq, qual, mima_loci, frameshift_loci)

    simple_read = self._SimpleRead(
      seq, qual, self.__get_five_prime__(pattern), pattern, mima_loci, n_loci, avg_qual
    )

    # self.mima_logic.print_mima(seq, qual, pattern)

    return simple_read

  def __get_phred_score__(self, seq, qual, mima_loci, frameshift_loci):
    if len(mima_loci) + len(frameshift_loci) == 0:
      return 0, 0

    remove_mimas = []
    phreds = []

    for start, end in frameshift_loci:
      fuse_phreds = []

      for i in range(start, end):
        if i in mima_loci:
          remove_mimas.append(i)
        fuse_phreds.append(ord(qual[i]) - self._phred_offset)

      phreds.append(min(fuse_phreds))

    for loci in mima_loci:
      if loci not in remove_mimas:
        phreds.append(ord(qual[loci]) - self._phred_offset)

    return np.mean(phreds), len(phreds)

  def __get_average_qual__(self, qual, mima_loci):
    if len(mima_loci) == 0:
      return 0
    phreds = np.array([ord(qual[i]) - self._phred_offset for i in mima_loci])
    return np.mean(phreds)

  def __trim_seq__(self, seq, qual):
    cutoff = 0
    min_sequence = 0
    for q in qual:
      qual_byte = ord(q) - self._phred_offset
      if qual_byte == 0:
        min_sequence += 1
      else:
        min_sequence = 0

      if min_sequence == 5:
        cutoff = cutoff - 5
        break

      cutoff += 1

    return seq[:cutoff], qual[:cutoff]

  def __get_five_prime__(self, pattern):
    if pattern is None:
      return None
    else:
      return pattern == "CCCTAA"

  def __get_pattern__(self, seq):
    cta, tag = "CCCTAA", "TTAGGG"
    pattern = None
    if cta in seq or tag in seq:
      if seq.count(cta) > seq.count(tag):
        pattern = cta
      else:
        pattern = tag
    return pattern

  def __flip_and_compliment__(self, read):
    if read.is_reverse:
      compliments = self._compliments
      seq_compliment = [compliments[base] for base in read.seq]
      seq_compliment = "".join(seq_compliment)
      return (seq_compliment[::-1], read.qual[::-1])
    else:
      return (read.seq, read.qual)


class MismatchingLociLogic(object):
  def get_telo_mismatch(self, seq, qual):

    c_rich_count = seq.count("CCCTAA")
    g_rich_count = seq.count("TTAGGG")

    if c_rich_count > g_rich_count:
      return self.get_mismatching_loci(seq, qual, "CCCTAA"), "CCCTAA"
    elif g_rich_count > c_rich_count:
      return self.get_mismatching_loci(seq, qual, "TTAGGG"), "TTAGGG"
    else:
      c_mima, c_fuse = self.get_mismatching_loci(seq, qual, "CCCTAA")
      g_mima, g_fuse = self.get_mismatching_loci(seq, qual, "TTAGGG")

      c_score = len(c_mima) + len(c_fuse)
      g_score = len(g_mima) + len(g_fuse)

      if c_score < g_score:
        return (c_mima, c_fuse), "CCCTAA"
      else:
        return (g_mima, g_fuse), "TTAGGG"

  def get_mismatching_loci(self, seq, qual, pattern):
    segments = re.split("(%s)" % (pattern,), seq)

    segments = self.join_complete_segments(segments, pattern)

    segment_offsets = self.get_segment_offsets(segments)
    self.extend_offsets(segments, pattern, segment_offsets)

    mima_loci, fuse_loci = self.offsets_to_loci(seq, qual, pattern, segment_offsets)

    return mima_loci, fuse_loci

  def join_complete_segments(self, segments, pattern):

    new_segements = []
    current_segment = ""

    for segment in segments:
      if segment == "":
        continue
      elif segment == pattern:
        current_segment += pattern
      else:
        if len(current_segment) > 0:
          new_segements.append(current_segment)
          current_segment = ""
        new_segements.append(segment)

    if len(current_segment) > 0:
      new_segements.append(current_segment)
    return new_segements

  def offsets_to_loci(self, seq, qual, pattern, segment_offsets):

    mima_loci = []
    fuse_loci = []
    seq_len = len(seq)

    # print segment_offsets
    # print [seq[s:e] for s, e in segment_offsets]

    for start, end in segment_offsets:
      segment = seq[start:end]
      segment_qual = qual[start:end]

      if start > 0 and start < end:
        # deletion event
        # +1 for exclusive upperrange
        end_of_range = min((start + 1, seq_len,))
        fuse_loci.append((start - 1, end_of_range))

      if pattern not in segment:
        if start > end:
          # These are fused loci
          fuse_loci.append((end, start))

        elif start < end:
          if (end - start) > 1:
            segment_mima = self.compare_to_telo(segment, segment_qual, pattern)

            segment_mima = [s + start for s in segment_mima]
            mima_loci.extend(segment_mima)
          else:
            mima_loci.append(start)

        elif start == end:
          # complete merge
          pass

    fuse_loci = self.filter_fuse_loci(mima_loci, fuse_loci)
    return mima_loci, fuse_loci

  def filter_fuse_loci(self, mima_loci, fuse_loci):
    remove_candidates = []
    offset = 0
    for loci in mima_loci:
      add_to_offset = 0
      for fuse_id, (start, end) in enumerate(fuse_loci[offset:]):
        if start <= loci and loci < end:
          remove_candidates.append(offset + fuse_id)
        elif loci > end:
          add_to_offset += 1
        elif loci < start:
          offset += add_to_offset
          break

    filtered_fuse = []
    for fuse_id, fuse in enumerate(fuse_loci):
      if fuse_id not in remove_candidates:
        filtered_fuse.append(fuse)

    all_fuse_loci = []
    for start, end in fuse_loci:
      all_fuse_loci.extend(range(start, end))

    merged_fuse = []
    current_fuse = []
    for i in set(all_fuse_loci):
      current_fuse.append(i)
      if len(current_fuse) > 1 and (current_fuse[-1] - current_fuse[-2]) > 1:
        merged_fuse.append((current_fuse[0], current_fuse[-1] + 1,))
        current_fuse = []

    return filtered_fuse

  def compare_to_telo(self, seq, qual, pattern):
    comparisons = []
    best_score = float("inf")

    for i in range(len(pattern)):
      comparison_seq = self.telo_sequence_generator(pattern, i)
      mima_loci = []
      qual_bytes = []
      score = 0
      for s, (l, c) in zip(seq, comparison_seq):
        if s != c:
          mima_loci.append(l)
          qual_bytes.append(ord(qual[l]))

          score += 1
          if score > best_score:
            break

      if score <= best_score:
        best_score = score

        if len(mima_loci) == 0:
          avg_phred = 0
        else:
          avg_phred = np.mean(qual_bytes)
        comparisons.append((score, list(mima_loci), avg_phred,))

    return self.get_best_offset(best_score, comparisons)

  def get_best_offset(self, best_score, comparisons):
    best_comparisons = []
    for score, loci, avg_phred in comparisons:
      if score == best_score:
        best_comparisons.append((avg_phred, loci))
    best_comparisons.sort(key=lambda x: x[0])
    return best_comparisons[0][1]

  def extend_offsets(self, segments, pattern, segment_offsets):
    for seg_id, segment in enumerate(segments):
      if pattern in segment:

        cur_seg_offsets = segment_offsets[seg_id]

        if seg_id > 0:
          # extend_backwards
          prev_seg_offsets = segment_offsets[seg_id - 1]

          new_offset = self.compare_to_pattern(
            segments[seg_id - 1], pattern, reverse=True
          )

          prev_seg_offsets[1] = prev_seg_offsets[1] - new_offset
          cur_seg_offsets[0] = cur_seg_offsets[0] - new_offset

        if seg_id < (len(segments) - 1):
          # extend_forwards
          next_seg_offsets = segment_offsets[seg_id + 1]
          new_offset = self.compare_to_pattern(segments[seg_id + 1], pattern)

          next_seg_offsets[0] = next_seg_offsets[0] + new_offset
          cur_seg_offsets[1] = cur_seg_offsets[1] + new_offset

  def get_segment_offsets(self, segments):
    segment_offsets = []
    offset = 0

    for segment in segments:
      segment_offsets.append(
        [offset, offset + len(segment),]
      )
      offset += len(segment)

    return segment_offsets

  def telo_sequence_generator(self, pattern, offset):

    offset_pattern = (pattern[offset:] + pattern)[: len(pattern)]
    i = 0
    while True:
      yield i, offset_pattern[i % len(pattern)]
      i += 1

  def compare_to_pattern(self, seq, pattern, reverse=False):
    i = 0
    generator = self.get_sequence_generator(seq, pattern, reverse)

    for c, s in generator:
      if c == s:
        i += 1
      else:
        break

    return i

  def get_sequence_generator(self, seq, pattern, reverse):
    def forward_gen(seq, pattern):
      for c, s in zip(seq, pattern):
        yield c, s
      return

    def reverse_gen(seq, pattern):
      for c, s in zip(seq[::-1], pattern[::-1]):
        yield c, s
      return

    if reverse:
      generator = reverse_gen(seq, pattern)
    else:
      generator = forward_gen(seq, pattern)

    return generator

  def print_mima(self, seq, qual, pat):
    print("-")
    loci_status, mima_loci, fuse_loci = self.get_loci_status(seq, qual, pat)

    print("Mima: %s" % str(mima_loci))
    print("Fuse: %s", str(fuse_loci))
    print(seq)
    print(loci_status)
    print(qual)

  def get_loci_status(self, seq, qual, pat):
    mima_loci, fuse_loci = self.get_mismatching_loci(seq, qual, pat)
    loci_status = []

    for i in range(len(seq)):
      if i in mima_loci:
        loci_status.append("X")
      else:
        loci_status.append("_")

    for start, end in fuse_loci:
      loci_status[start:end] = ["F"] * (end - start)

    return "".join(loci_status), mima_loci, fuse_loci


class VitalStatsFinder(object):
  def __init__(self, temp_dir, total_procs, task_size, trim_length=0):
    self.temp_dir = temp_dir
    self._total_procs = total_procs
    self._task_size = task_size
    self._trim_length = trim_length

  def __csv_to_dict__(self, stats_path):
    insert_dat = pd.read_csv(stats_path).to_dict(orient="records")[0]

    ins_N = int(insert_dat["N"])
    if ins_N == 0:
      insert_mean = -1
      insert_sd = -1
    else:
      ins_sum = int(insert_dat["sum"])
      ins_power_2 = int(insert_dat["power_2"])

      insert_mean, insert_sd = self.__get_mean_and_sd__(ins_sum, ins_power_2, ins_N)

    min_qual = int(insert_dat["min_qual"])
    qual_mean, qual_sd = self.__get_mean_and_sd__(
      insert_dat["qual_sum"], insert_dat["qual_power_2"], insert_dat["qual_N"]
    )

    return {
      "insert_mean": insert_mean,
      "insert_sd": insert_sd,
      "min_qual": min_qual,
      "max_qual": int(insert_dat["max_qual"]),
      "read_len": int(insert_dat["read_len"]),
      "qual_mean": qual_mean,
      "qual_sd": qual_sd,
    }

  def __get_mean_and_sd__(self, x_sum, x_power_2, x_N):

    x_mean = x_sum / x_N
    x_sd = np.sqrt(float((x_N * x_power_2)) - float(x_sum ** 2)) / x_N

    return x_mean, x_sd

  def get_vital_stats(self, sample_path):

    vital_stats_csv = self.__run_vital_rule__(sample_path)
    vital_stats = self.__csv_to_dict__(vital_stats_csv)
    vital_stats["phred_offset"] = vital_stats["min_qual"]
    vital_stats["initial_read_len"] = vital_stats["read_len"]

    if self._trim_length > 0:
      vital_stats["read_len"] = self._trim_length

    return vital_stats

  def __run_vital_rule__(self, sample_path, keep_in_temp=True):
    def rule(read, constants, master):
      stats = {}

      if read.is_read1 and read.is_proper_pair and read.mapq > 38:
        insert_size = abs(read.template_length)
        stats["sum"] = insert_size
        stats["power_2"] = insert_size ** 2
        stats["N"] = 1

      stats["read_len"] = len(read.seq)
      byte_vals = [ ord(char) for char in read.qual ]

      min_qual = min(byte_vals)
      max_qual = max(byte_vals)

      qual_mean = np.mean(byte_vals)
      stats["qual_sum"] = qual_mean
      stats["qual_power_2"] = qual_mean ** 2
      stats["qual_N"] = 1

      stats["min_qual"] = min_qual
      stats["max_qual"] = max_qual

      return stats

    structures = {}

    structures["sum"] = {"data": 0, "store_method": "cumu"}
    structures["power_2"] = {"data": 0, "store_method": "cumu"}
    structures["N"] = {"data": 0, "store_method": "cumu"}
    structures["read_len"] = {"data": 0, "store_method": "max"}

    structures["min_qual"] = {"data": 999, "store_method": "min"}
    structures["max_qual"] = {"data": 0, "store_method": "max"}

    structures["qual_sum"] = {"data": 0, "store_method": "cumu"}
    structures["qual_power_2"] = {"data": 0, "store_method": "cumu"}
    structures["qual_N"] = {"data": 0, "store_method": "cumu"}

    stat_interface = parabam.Stat(
      temp_dir=self.temp_dir,
      total_procs=self._total_procs,
      task_size=10000,
      keep_in_temp=keep_in_temp,
    )

    out_paths = stat_interface.run(
      input_paths=[sample_path], constants={}, rule=rule, struc_blueprint=structures
    )

    return out_paths["global"]["stats"]


class ReadStatsFactory(object):
  def __init__(self, temp_dir, total_procs=4, task_size=5000, trim_reads=0, seed_randomness=False, debug_print=False):

    self.temp_dir = temp_dir
    self._total_procs = total_procs
    self._task_size = task_size

    self._debug_print = debug_print
    self._trim = trim_reads
    self.seed_randomness = seed_randomness

  def get_read_counts(self, path, vital_stats):
    read_stat_paths = self.run_read_stat_rule(path, vital_stats)

    read_array = self.__path_to_read_array__(read_stat_paths["read_array"])

    error_profile, sample_variance = self.__paths_to_error_profile__(read_stat_paths)

    read_counts = self.read_array_to_counts(read_array, error_profile, sample_variance)

    self.__delete_analysis_paths__(read_stat_paths)

    return read_counts

  def __delete_analysis_paths__(self, read_stat_paths):
    for analysis, path in read_stat_paths.items():
      os.remove(path)

  def __paths_to_error_profile__(self, read_stat_paths):
    random_counts = pd.read_csv(read_stat_paths["random_counts"], header=None).values
    read_counts = pd.read_csv(read_stat_paths["mima_counts"], header=None).values
    error_profile = self.get_error_profile(read_counts, random_counts)
    sample_variance = self.__get_sample_variance__(read_counts)

    return error_profile, sample_variance

  def __get_sample_variance__(self, read_counts):
    read_counts[0, :] = 0
    mask = read_counts > 0
    mask[40:, :] = False

    read_variance = read_counts[mask].std() / read_counts[mask].mean()
    return read_variance

  def get_error_profile(self, read_counts, random_counts, thresh=None):
    error_profile = self.__get_significantly_enriched__(read_counts, random_counts, thresh)

    error_profile = self.__remove_noise__(error_profile)
    error_profile = self.__prune_error_profile__(error_profile)
    error_profile = self.__rationalise_error_profile__(error_profile)

    error_profile[: int(read_counts.shape[0] * 0.1), :] = 1

    return error_profile

  def __get_significantly_enriched__(self, read_counts, random_counts, thresh=None):
    dif_counts = read_counts - random_counts
    ten_percent = int(read_counts.shape[0] * 0.1)

    if thresh is None:
      mask = self.__get_exclusion_mask__(read_counts)
      arg_max_index = (read_counts * mask).argmax()
      dif_loci_x, dif_loci_y = np.unravel_index(arg_max_index, dif_counts.shape)

      hi_thresh = dif_counts[
        int(dif_loci_x - ten_percent) : int(dif_loci_x + ten_percent),
        dif_loci_y - 15 : dif_loci_y + 1,
      ]
      hi_thresh = hi_thresh.flatten()

      thresh = np.percentile(hi_thresh, 95)

    if self._debug_print:
      print("Thresh:", thresh)

    error_profile = (dif_counts * (dif_counts > 0)) > thresh
    error_profile = error_profile * 1

    return error_profile

  def __remove_noise__(self, error_profile):
    row_max, col_max = error_profile.shape
    error_profile[int(row_max * 0.11) :, int(col_max * 0.7) :] = 0
    error_profile[int(row_max * 0.40) :, 0] = 0
    error_profile[int(row_max * 0.55) :, :] = 0

    return error_profile

  def __get_exclusion_mask__(self, read_counts):
    mask = np.zeros(read_counts.shape)
    x_start = int(read_counts.shape[0] * 0.2)
    y_start = int(read_counts.shape[1] * 0.5)
    mask[x_start:, y_start:] = 1
    return mask

  def __prune_error_profile__(self, error_profile):

    isolated_mask = self.__get_isolated_mask__(error_profile)
    continuous_mask = self.__get_continuous_mask__(error_profile)

    combined_mask = (isolated_mask + continuous_mask) > 0

    return error_profile * combined_mask

  def __get_continuous_mask__(self, error_profile):
    row_continuous = self.__transform_continous_matrix__(error_profile)
    col_continuous = self.__transform_continous_matrix__(error_profile.transpose())
    col_continuous = col_continuous.transpose()

    max_continuous = row_continuous * (row_continuous > col_continuous)
    max_continuous = max_continuous + (col_continuous * (col_continuous >= row_continuous))

    continusous_mask = max_continuous >= 4
    return continusous_mask

  def __transform_continous_matrix__(self, error_profile):
    continuous = np.zeros(error_profile.shape)

    for row_i in range(0, error_profile.shape[0]):
      sequence = 0
      start_index = -1
      for col_i in range(0, error_profile.shape[1]):
        value = error_profile[row_i, col_i]
        if value == 0:
          if sequence > 0:
            continuous[row_i, start_index:col_i] = sequence
            sequence = 0
            start_index = -1
        elif value > 0:
          if start_index == -1:
            start_index = col_i
          sequence += 1
    return continuous

  def __get_isolated_mask__(self, error_profile):
    isolated_mask = np.ones(error_profile.shape)
    first_locis = self.__get_first_loci__(error_profile)
    for row_i in range(1, error_profile.shape[0]):
      if first_locis[row_i] == -1:
        # skip rows with no entries
        continue
      else:
        for col_i in range(error_profile.shape[1]):
          if (not error_profile[row_i, col_i]) or col_i == 0:
            continue
          elif self.__prune_decision__(row_i, col_i, error_profile):
            isolated_mask[row_i, col_i] = 0
    return isolated_mask

  def __prune_decision__(self, row_i, col_i, error_profile):
    neighbours = [
      (row_i - 1, col_i + 1),
      (row_i - 1, col_i),
      (row_i - 1, col_i - 1),
      (row_i, col_i + 1),
      (row_i, col_i - 1),
      (row_i + 1, col_i + 1),
      (row_i + 1, col_i),
      (row_i + 1, col_i - 1),
    ]

    try:
      return self.__get_neighbor_sum__(row_i, col_i, error_profile, neighbours) < 4
    except IndexError:
      return False

  def __get_neighbor_sum__(self, row_i, col_i, error_profile, neighbours):
    neighbours_sum = sum([error_profile[r, c] for (r, c) in neighbours])
    return neighbours_sum

  def __get_first_loci__(self, error_profile):
    first_loci = []
    for row_i in range(error_profile.shape[0]):
      if any(error_profile[row_i, :]):
        for col_i in range(error_profile.shape[1]):
          if error_profile[row_i, col_i]:
            first_loci.append(col_i)
            break
      else:
        first_loci.append(-1)
    return first_loci

  def __rationalise_error_profile__(self, error_profile):
    if error_profile.sum() > 0:
      start_row = np.where(error_profile)[0].max()
      global_loci = 0
      for i in reversed(range(0, start_row + 1)):
        error_bins_in_row = np.where(error_profile[i, :])[0]
        if len(error_bins_in_row) > 0:
          cur_loci = error_bins_in_row.max()
        else:
          cur_loci = 0

        # if cur_loci > global_loci:
        global_loci = cur_loci
        error_profile[i, : global_loci + 1] = True

    return error_profile

  def __array_to_file__(self, array, unique):
    df = pd.DataFrame(array)
    out_path = "./%s-tmctout.csv" % (unique)
    df.to_csv(out_path, index=False, header=False)
    return out_path

  def __path_to_read_array__(self, read_array_path):
    return pd.read_csv(read_array_path, header=None).values

  def read_array_to_counts(self, read_array, error_profile, sample_variance):
    complete_reads, boundary_reads = self.__get_complete_status__(read_array, error_profile)

    f2_count, f4_count = self.__get_boundary_counts__(boundary_reads)
    f1_count = self.__get_f1_count__(complete_reads)

    return_dat = {
      "F2": int(f2_count),
      "F1": int(f1_count),
      "F4": f4_count,
      "sample_variance": sample_variance,
    }

    return return_dat

  def __get_f1_count__(self, complete_reads):
    return float(complete_reads.shape[0]) / 2

  def __get_boundary_counts__(self, boundary_reads):
    f2_count, f4_count, total_reads = self.__get_read_counts__(boundary_reads)
    return f2_count, f4_count

  def __get_read_counts__(self, boundary_reads):
    f2_count = sum(boundary_reads[:, 3] == 1)
    f4_count = sum(boundary_reads[:, 3] == 0)
    total_reads = boundary_reads.shape[0]
    return f2_count, f4_count, total_reads

  def __get_complete_status__(self, read_array, error_profile):
    boundary_indicies = []
    complete_indicies = []

    for i in range(int(read_array.shape[0])):
      read_info = list(map(int, read_array[i, [0, -2]]))
      pair_info = list(map(int, read_array[i, [2, -1]]))

      read = error_profile[read_info[0], read_info[1]]
      pair = error_profile[pair_info[0], pair_info[1]]

      if read and pair:
        complete_indicies.append(i)
      elif (not read) and pair:
        boundary_indicies.append(i)

    return read_array[complete_indicies, :], read_array[boundary_indicies, :]

  def run_read_stat_rule(self, path, vital_stats, keep_in_temp=True):

    simple_read_factory = SimpleReadFactory(vital_stats, trim_reads=self._trim)
    phred_offset = vital_stats["phred_offset"]

    maxtrix_max = (vital_stats["max_qual"] - phred_offset) + 1
    matrix_shape = (vital_stats["read_len"] + 1, maxtrix_max)

    def get_return_stats(reads):

      return_stats = [
        len(reads[0].mima_loci),
        int(reads[0].five_prime),
        len(reads[1].mima_loci),
        int(reads[1].five_prime),
        reads[0].avg_qual,
        reads[1].avg_qual,
      ]

      return return_stats

    def rule(reads, constants, master, seed_randomness):
      simple_reads = [simple_read_factory.get_simple_read(read) for read in reads]
      return_dat = np.zeros((2, 6))
      return_dat[0, :] = get_return_stats(simple_reads)
      return_dat[1, :] = get_return_stats(simple_reads[::-1])

      random_counts = np.zeros(matrix_shape)
      mima_counts = np.zeros(matrix_shape)

      if seed_randomness:
          random.seed(RANDOM_SEED)

      for read in simple_reads:
        mima_counts[read.n_loci, int(read.avg_qual)] += 1

        sample_size = len(read.mima_loci)
        if sample_size > 0:
          rand_quals = random.sample(list(read.qual), sample_size)
          qual_bytes = [ord(q) - phred_offset for q in rand_quals]
          rand_avg = np.mean(qual_bytes)

          random_counts[int(sample_size), int(rand_avg)] += 1

      results = {
        "read_array": np.array(return_dat),
        "random_counts": random_counts,
        "mima_counts": mima_counts,
      }

      return results

    structures = {
      "read_array": {"data": np.zeros((2, 6)), "store_method": "vstack"},
      "mima_counts": {"data": np.zeros(matrix_shape), "store_method": "cumu"},
      "random_counts": {"data": np.zeros(matrix_shape), "store_method": "cumu"},
    }

    stat_interface = parabam.Stat(
      temp_dir=self.temp_dir,
      pair_process=True,
      total_procs=self._total_procs,
      task_size=self._task_size,
      keep_in_temp=keep_in_temp,
      verbose=0,
    )

    out_paths = stat_interface.run(
      input_paths=[path], constants={}, rule=partial(rule, seed_randomness=self.seed_randomness), struc_blueprint=structures
    )

    return out_paths[path]


class Telbam2Length(TelomerecatInterface):
  def __init__(
    self,
    temp_dir=None,
    task_size=10000,
    total_procs=4,
    reader_n=2,
    verbose=False,
    announce=True,
    cmd_run=False,
  ):

    super(Telbam2Length, self).__init__(
      instance_name="telomerecat telbam2length",
      temp_dir=temp_dir,
      task_size=task_size,
      total_procs=total_procs,
      reader_n=reader_n,
      verbose=verbose,
      announce=announce,
      cmd_run=cmd_run,
    )

  def run_cmd(self):
    self.run(
      input_paths=self.cmd_args.input,
      trim=self.cmd_args.trim,
      output_path=self.cmd_args.output,
      simulator_n=self.cmd_args.simulator_runs,
      correct_f2a=self.cmd_args.enable_correction,
      inserts_path=self.cmd_args.insert,
      seed_randomness=self.cmd_args.seed_randomness
    )

  def run(
    self,
    input_paths,
    trim=0,
    output_path=None,
    correct_f2a=False,
    simulator_n=10,
    inserts_path=None,
    seed_randomness=False
  ):

    """The main function for invoking the part of the
       program which creates a telbam from a bam

    Arguments:
      inputs_paths (list): The TELBAMs that we wish to estimate TL estimates for
      output_path (string): Specify a path to output results to (optional)
      inserts_path (string): A path to a file containing insert length estimates
                   for each TELBAM. Formatted as follows:
                    example_telbam.bam, insert_mean, insert_sd
    """

    self.__introduce__()
    names = [ os.path.basename(path).replace("_telbam", "") for path in input_paths ]

    output_csv_path = output_path
    temp_csv_path = self.__get_temp_path__()

    insert_length_generator = self.__get_insert_generator__(inserts_path)

    self.__output__(f" Collecting meta-data for all samples | {self.__get_date_time__()}", 1)

    vital_stats_finder = VitalStatsFinder(
      self.temp_dir, self.total_procs, self.task_size, trim
    )

    for sample_path, sample_name, in zip(input_paths, names):
      self.__output__(f"\t- {sample_name} | {self.__get_date_time__()}", 2)

      vital_stats = vital_stats_finder.get_vital_stats(sample_path)

      self.__check_vital_stats_insert_size__(
        inserts_path, insert_length_generator, vital_stats
      )

      read_type_counts = self.__get_read_types__(
        sample_path, vital_stats, self.total_procs, trim, seed_randomness
      )

      self.__write_to_csv__(read_type_counts, vital_stats, temp_csv_path, sample_name)

    self.__output__('', 1)
    length_interface = Csv2Length(
      temp_dir=self.temp_dir,
      total_procs=self.total_procs,
      verbose=self.verbose,
      announce=False,
      cmd_run=False,
    )

    length_interface.run(
      input_paths=[temp_csv_path],
      output_paths=[output_csv_path],
      correct_f2a=correct_f2a,
      seed_randomness=seed_randomness,
      simulator_n=simulator_n
    )

    self.__print_output_information__(output_csv_path)
    self.__goodbye__()

    return output_csv_path

  def __print_output_information__(self, output_csv_path):
    self.__output__(" Length estimation results written to the following file:", 1)
    self.__output__(f"\t./{os.path.basename(output_csv_path)}\n", 1)  # extra line feed

  def __get_insert_generator__(self, inserts_path):
    if inserts_path:
      with open(inserts_path, "r") as inserts_file:
        for line in inserts_file:
          yield map(float, line.split(","))

  def __check_vital_stats_insert_size__(
    self, inserts_path, insert_length_generator, vital_stats
  ):
    if inserts_path:
      insert_mean, insert_sd = insert_length_generator.__next__()
      vital_stats["insert_mean"] = insert_mean
      vital_stats["insert_sd"] = insert_sd
      self.__output__(f"\t\t+ Using user defined insert size: {int(insert_mean)},{int(insert_sd)}", 2)
    elif vital_stats["insert_mean"] == -1:
      default_mean, default_sd = 350, 25
      vital_stats["insert_mean"] = 350
      vital_stats["insert_sd"] = 25
      self.__output__(f"\t\t+ Failed to estimate insert size. Using default: {int(default_mean)},{int(default_sd)}", 2)

  def __get_read_types__(
    self, sample_path, vital_stats, total_procs, trim, seed_randomness, read_stats_factory=None
  ):

    if read_stats_factory is None:
      read_stats_factory = ReadStatsFactory(
        temp_dir=self.temp_dir, total_procs=total_procs, trim_reads=trim, seed_randomness=seed_randomness, debug_print=False
      )

    read_type_counts = read_stats_factory.get_read_counts(sample_path, vital_stats)
    return read_type_counts

  def __get_temp_path__(self):
    temp_path = os.path.join(self.temp_dir, "telomerecat_temp_%d.csv" % (time.time()))
    self.__create_output_file__(temp_path)
    return temp_path

  def __create_output_file__(self, output_csv_path):
    with open(output_csv_path, "w") as total:
      header = (
        "Sample,F1,F2,F4,Psi,Insert_mean,Insert_sd," "Read_length,Initial_read_length\n"
      )
      total.write(header)
    return output_csv_path

  def __write_to_csv__(self, read_type_counts, vital_stats, output_csv_path, name):
    with open(output_csv_path, "a") as counts:
      counts.write(
        "%s,%d,%d,%d,%.3f,%.3f,%.3f,%d,%d\n"
        % (
          name,
          read_type_counts["F1"],
          read_type_counts["F2"],
          read_type_counts["F4"],
          read_type_counts["sample_variance"],
          vital_stats["insert_mean"],
          vital_stats["insert_sd"],
          vital_stats["read_len"],
          vital_stats["initial_read_len"],
        )
      )

  def get_parser(self):
    parser = self.default_parser()
    parser.description = textwrap.dedent(
      """\
    %s
    %s

      The telbam2length command allows the user to genereate a telomere
      length estimate from a previously generated TELBAM file.

      Example useage:

      telomerecat telbam2length /path/to/some_telbam.bam

      This will generate a .csv file with an telomere length estimate
      for the `some_telbam.bam` file.

    %s
    """
      % (self.instance_name, self.header_line, self.header_line,)
    )

    for arg_name in ['input_telbam', 'output_csv', 'trim', 'nreads_for_task']:
      add_arg[arg_name](parser)

    return parser


if __name__ == "__main__":
  print("Do not run this script directly. Type `telomerecat` for help.")



from typing import List
import os
import sys
import pysam
import logging
import tempfile
import multiprocessing as mp
import subprocess
from functools import partial

from telomerecat.constants import TEL_PATS, HTS_EXT_TO_AF_MODE
from pysam import AlignmentFile

def _log_setup(loglevel):
    logging.basicConfig(level=getattr(logging, loglevel.upper()), format='%(levelname)s: %(message)s')


def get_align_file(path:str, mode='r', template=None, expectIndex=True, threads=1):
    hts_ext = os.path.splitext(path)[-1]
    xam_type = HTS_EXT_TO_AF_MODE[hts_ext]
    if expectIndex is False:
        save = pysam.set_verbosity(0)
    af = pysam.AlignmentFile(path, f'{mode}{xam_type}', template=template, threads=threads)
    if expectIndex is False:
        pysam.set_verbosity(save)
    return af


def get_processes(processes):
    if processes == -1:
        processes = len(os.sched_getaffinity(0))
    return processes


def get_hts_processes(processes):
    if processes > 2:
        processes = 2
    return processes


def telbam_path(dest_dir:str, xam_file:str):
    (base, _) = os.path.splitext(os.path.basename(xam_file))
    return os.path.join(dest_dir, f'{base}_telbam.bam')


def collate_pairs(xam_file: str, tmpdir:str, processes=1, reference=None):
    pairs = os.path.join(tmpdir, 'pairs.bam')
    collate_opts = ['--no-PG', '-f', '-n', '1', '-l', '1', '-r', '5000000']  # ~8GB RAM
    if processes > 1:
        collate_opts.extend(['-@', str(processes)])
    if reference:
        collate_opts.extend(['--reference', reference])
    collate_opts.extend(['-o', pairs, xam_file, os.path.join(tmpdir, 'collate')])
    pysam.collate(*collate_opts) # don't forget to splat, this isn't perl
    return pairs


def pairs_to_telbam(af_pairs:AlignmentFile, af_telbam:AlignmentFile):
    read_iter = af_pairs.fetch(until_eof=True)
    while True:
        read_a = next(read_iter, None)
        if read_a is None:
            break
        read_b = next(read_iter)
        qseq = read_a.query_sequence
        if TEL_PATS[0] in qseq or TEL_PATS[1] in qseq:
            af_telbam.write(read_a)
            af_telbam.write(read_b)
        else:
            qseq = read_b.query_sequence
            if TEL_PATS[0] in qseq or TEL_PATS[1] in qseq:
                af_telbam.write(read_a)
                af_telbam.write(read_b)
    return


def to_telbam(xam_file:str, outbam_dir:str, tmpdir:str, hts_processes=1, reference=None):
    telbam = telbam_path(outbam_dir, xam_file)
    telbam_paths = {xam_file: {'telbam': telbam}}

    with tempfile.TemporaryDirectory(prefix='telomerecat_', dir=tmpdir) as this_tmp:
        collate_wrap = ['pysam_collate', '-t', str(this_tmp), '-@', str(hts_processes)]
        if reference:
            collate_wrap.extend(['--reference', reference])
        collate_wrap.append(xam_file)
        named_fifo = os.path.join(this_tmp, 'collatepipe.bam')
        os.mkfifo(named_fifo)
        collate_wrap.extend(['--output', named_fifo])
        collate_proc = subprocess.Popen(collate_wrap)
        af_pairs = get_align_file(named_fifo, mode='r', expectIndex=False, threads=hts_processes)
        af_telbam = get_align_file(telbam, mode='w', template=af_pairs, threads=hts_processes)
        pairs_to_telbam(af_pairs, af_telbam)

        # we should be finished anyway
        (out_proc, err_proc) = collate_proc.communicate(timeout=15)
        collate_exit_code = collate_proc.returncode
        if collate_exit_code != 0:
            print(f'ERROR: samtools collate process exited with a non-zero value ({collate_exit_code})')
            sys.exit(collate_exit_code)

    return telbam_paths


def process_alignments(outbam_dir:str, processes:int, tmpdir:str, files:List[str], reference=None, verbose=0):
    if verbose > 0:
        # for compatibility with original telomerecat
        _log_setup('DEBUG')
    telbam_paths = {}
    # general max CPUs (works correctly for cgroups)
    processes = get_processes(processes)
    # max threads for HTS actions
    hts_processes = get_hts_processes(processes)

    results = []
    ctx = mp.get_context('forkserver')
    logging.debug(f'Starting telbam generation for {len(files)} inputs')

    parallel_processes = int(processes/hts_processes)
    # allow non-multiprocess use when minimal files, helpful for debugging
    if parallel_processes <= 1 or len(files) == 1:
        logging.debug(f'Inline execution of telbam generation, {hts_processes} threads per process')
        # allow max threads for helpers
        hts_processes = processes
        for xam_file in files:
            telbam_path = to_telbam(xam_file, outbam_dir=outbam_dir, tmpdir=tmpdir, hts_processes=hts_processes, reference=reference)
            telbam_paths.update(telbam_path)
    else:
        logging.debug(f'Multiprocessing execution of telbam generation, {hts_processes} threads per process, {parallel_processes} in parallel')
        with ctx.Pool(processes=parallel_processes) as pool:
            results = pool.map(partial(to_telbam, outbam_dir=outbam_dir, tmpdir=tmpdir, hts_processes=hts_processes, reference=reference), files)
        for telbam_path in results:
            telbam_paths.update(telbam_path)

    return telbam_paths



#!/usr/bin/env python3
import os
import pkg_resources
import pysam
import click

@click.command()
@click.version_option(pkg_resources.require("telomerecat")[0].version)
@click.option('--output', required=True, type=str, help='File or named pipe')
@click.option('-@', '--threads', required=False, type=int, default=1, show_default=True, help='Helper threads')
@click.option('--reference', required=False, type=str, default=None, show_default=True, help='Reference FASTA, required for CRAM')
@click.option('-t', '--tmpdir', required=True, type=click.Path(exists=True, resolve_path=True), help='Large preexisting tempdir')
@click.argument('in_bam_cram', nargs=1, required=True, type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, resolve_path=True))
def thin_wrap(output, threads, reference, tmpdir, in_bam_cram):
    """Simple wrapper around pysam.collate to avoid requiring system installed samtools"""

    prefix = os.path.join(tmpdir, 'collate')
    # although this is intended to be used via a named pipe this will reduce the chance of it being filled
    collate_opts = ['--no-PG', '-r', '2500000', '-f', '--output-fmt', 'BAM', '-l', '1']
    collate_opts.extend(['-o', output])
    collate_opts.extend(['-@', str(threads)])
    if reference:
        collate_opts.extend(['--reference', reference])
    collate_opts.append(in_bam_cram)
    collate_opts.append(prefix)

    pysam.collate(*collate_opts)





#!/usr/bin/env python3

from setuptools import setup

setup(
  name="telomerecat",
  version='4.0.2',
  description="Telomere Computational Analysis Tool",
  long_description=open('README.md').read(),
  long_description_content_type='text/markdown',
  url='https://github.com/cancerit/telomerecat',
  author="JHR Farmery",
  license="GPL",
  python_requires='>= 3.7',
  author_email="cgphelp@sanger.ac.uk",
  packages=["telomerecat"],
  package_dir={"telomerecat": "telomerecat"},
  install_requires=["parabam>=3.0.1", "numpy", "pysam", "pandas", "click"],
  include_package_data=True,
  scripts=["./telomerecat/bin/telomerecat"],
  entry_points={'console_scripts': ['pysam_collate=telomerecat.pysam_collate:thin_wrap'],},
  zip_safe=False,
)



"""
Generate TL estimates given a CSV file.

Run the telomerecat length estimate on a CSV containing the
the following fields:
  F1, F2, F4, Psi, Insert_mean, Insert_sd

Author: jhrf
"""

import sys
import textwrap
import time
import random
import math

import numpy as np
import pandas as pd

from argparse import SUPPRESS
from functools import partial
from shutil import copy
from multiprocessing import Pool, freeze_support, get_context

from telomerecat import core

from . import RANDOM_SEED

class LengthSimulator(object):
  def __init__(self, insert_mu, insert_sigma, complete, boundary, read_len, seed_randomness=False):

    self._insert_mu = insert_mu
    self._insert_sigma = insert_sigma

    self._obs_total = complete + boundary  # observed total
    self._read_thresh = int(self._obs_total * 0.0001)

    self._complete = complete
    self._boundary = boundary
    self._read_len = read_len
    self._read_sim = self.__simulate_reads__

    self._get_read_len = lambda: self._read_len
    self.seed_randomness = seed_randomness

  def start(self):
    if self._boundary <= 0 or self._complete <= 0:
      return 0
    else:
      return self.__run_sim__()

  def __run_sim__(self):
    found = False

    tel_len = int((float(self._complete) / self._boundary) * (self._insert_mu - 100))
    max_its = 5000
    its = 0

    best_result = float("inf")
    best_tel_len = tel_len

    read_simulator = self._read_sim
    get_factor = self.__get_factor__

    while not found and its < max_its:
      sim_comp, sim_boun, invalid_reads = read_simulator(tel_len)
      #   result can be positive or negative thus
      #   influencing elongation and shortening
      #   A negative diference (overestimate) results
      #   in a shorter telomere length next iteration
      result = self.is_same_ratio(sim_comp, sim_boun)
      abs_result = abs(result)

      if abs_result < best_result:
        best_result = abs_result
        best_tel_len = tel_len

      if result == 0:
        found = True
      else:
        mult = -1 if result < 0 else 1
        factor = get_factor(abs_result)
        tel_len += factor * mult
      its += 1

    return best_tel_len

  def __get_factor__(self, abs_result):
    total = self._obs_total
    if self.seed_randomness:
      random.seed(RANDOM_SEED)
    if abs_result < (total * 0.001):
      factor = 1
    elif abs_result < (total * 0.2):
      factor = random.sample([1, 2, 2, 4, 25, 200, 500], 1)[0]
    elif abs_result < (total * 0.5):
      factor = random.sample([20, 50, 100, 200, 1000, 2000], 1)[0]
    else:
      factor = int(np.log(abs_result) * 300) / random.sample([1, 2], 1)[0]
    return factor

  def is_same_ratio(self, comp, boun):
    difference = abs(comp - self._complete)
    if difference <= self._read_thresh:
      return 0
    else:
      return self._complete - comp

  def __simulate_reads__(self, tel_len):
    if tel_len < self._insert_mu:
      return int(self._complete * 0.50), self._boundary, 0

    #  speedup
    is_complete = self.is_complete
    is_boundary = self.is_boundary

    insert_mean = self._insert_mu
    insert_sigma = self._insert_sigma

    obs_total = self._obs_total
    #  speedup

    est_complete = 0
    est_boundary = 0
    invalid_read = 0
    total = 0

    if self.seed_randomness:
      random.seed(RANDOM_SEED)

    tel_len_int = int(math.floor(tel_len))

    while total < (obs_total):
      insert_size = random.gauss(insert_mean, insert_sigma)
      location = random.randint(0, tel_len_int)
      if is_complete(location, insert_size):
        est_complete += 1
        total += 1
      elif is_boundary(location, insert_size):
        est_boundary += 1
        total += 1
      else:
        invalid_read += 1

    return (est_complete, est_boundary, invalid_read)

  def is_complete(self, location, insert_size):
    return (location - insert_size) > 0

  def is_boundary(self, location, insert_size):
    return (location - self._get_read_len()) > 0


class Csv2Length(core.TelomerecatInterface):
  def __init__(self, temp_dir=None, total_procs=4, verbose=False, announce=True, cmd_run=False):

    super(Csv2Length, self).__init__(
      instance_name="telomerecat csv2length",
      temp_dir=temp_dir,
      total_procs=total_procs,
      verbose=verbose,
      announce=announce,
      cmd_run=cmd_run,
    )

  def run_cmd(self):
    output_paths = self.__handle_cmd_outpaths__()

    self.run(
      input_paths=self.cmd_args.input,
      correct_f2a=self.cmd_args.enable_correction,
      output_paths=output_paths,
      prior_weight=self.cmd_args.prior_weight,
      seed_randomness=self.cmd_args.seed_randomness,
      simulator_n=self.cmd_args.simulator_runs
    )

  def run(self, input_paths, output_paths=[], correct_f2a=True, prior_weight=3, seed_randomness=False, simulator_n=10):

    self.__introduce__()
    self.__generate_output_paths__(input_paths, output_paths)

    self.__output__(f" Commencing length estimation | {self.__get_date_time__()}", 1)

    for input_path, output_path in zip(input_paths, output_paths):
      if self.announce:
        self.__output__(f"\tInput: {input_path}", 1)
        self.__output__(f" \tOutput: {output_path}", 1)

      counts = pd.read_csv(input_path)
      counts = self.__get_length_from_dataframe__(
        counts, simulator_n, correct_f2a, prior_weight, seed_randomness
      )

      self.__output_length_results__(counts, output_path)
      self.__output__('', 1)

    self.__goodbye__()
    return output_paths

  def __handle_cmd_outpaths__(self):
    output_paths = None
    if self.cmd_args.output is None:
      output_paths = []
    else:
      output_paths = self.cmd_args.output.split(",")
      output_paths = [p.strip() for p in output_paths]

    return output_paths

  def __get_length_from_dataframe__(
    self, counts, simulator_n, correct_f2a=True, prior_weight=3, seed_randomness=False, simulate_lengths=True,
  ):

    counts["F2a"] = counts["F2"] - counts["F4"]

    if correct_f2a:
      counts["F2a_c"] = self.__get_corrected_f2a__(counts, prior_weight)
    else:
      counts["F2a_c"] = counts["F2a"]

    if simulate_lengths:
      counts["Length"] = self.__get_lengths__(counts, seed_randomness, simulator_n)
    else:
      counts["Length"] = self.__quick_length__(counts)

    return counts

  def __get_corrected_f2a__(self, counts, prior_weight=3):

    theta_observed = counts["F2a"] / (counts["F2"] + counts["F4"])

    theta_expected = sum(theta_observed * counts["F2"]) / sum(counts["F2"])

    theta_corrected = (
      (theta_observed * (counts["Psi"])) + (theta_expected * prior_weight)
    ) / ((counts["Psi"]) + (prior_weight))

    corrected_f2_counts = (counts["F2"] + counts["F4"]) * theta_corrected

    return corrected_f2_counts.round(3)

  def __quick_length__(self, counts):
    lengths = []
    for i, sample in counts.iterrows():
      factor = sample["Read_length"] / (sample["F2a_c"] - sample["Read_length"])
      scale = sample["F2a_c"] + (sample["F2a_c"] * factor)
      length = ((sample["F1"] / scale) * sample["Insert_mean"]) + sample["Insert_mean"]
      lengths.append(round(length, 3))
    return lengths

  def __get_lengths__(self, counts, seed_randomness, simulator_n):
    lengths = []
    for i, sample in counts.iterrows():
      self.__output__(f"\t- {sample['Sample']} | {self.__get_date_time__()}", 2)

      length_mean, len_std = run_simulator_par(
        sample["Insert_mean"],
        sample["Insert_sd"],
        sample["F1"],
        sample["F2a_c"],
        self.total_procs,
        sample["Read_length"],
        seed_randomness,
        simulator_n
      )

      lengths.append(length_mean)
    return lengths

  def __generate_output_paths__(self, input_paths, output_paths):
    if len(output_paths) < len(input_paths):
      for i in range(len(input_paths)):
        path = "./telomerecat_length_%d_%d.csv" % (time.time(), i)
        output_paths.append(path)

  def __output_length_results__(self, counts, count_path):
    counts.to_csv(count_path, index=None)

  def __copy_out_of_temp__(self, file_paths, copy_path="."):
    for fil in file_paths:
      copy(fil, copy_path)

  def get_parser(self):
    parser = self.default_parser()
    parser.description = textwrap.dedent(
      """\
    %s
    %s

      The csv2length command allows the user to genereate new TL
      estimates from a previously generated CSV. This could be useful
      if the user wishes to re-run analysis with different options
      (such as F2a correction) and does not wish to reanalyse
      the TELBAM files.

      Example useage:

      telomerecat csv2length /path/to/telomerecat_length_123.csv

      This will generate a new .csv file with revised telomere length
      estimates for the `telomerecat_length_123.csv` file.

    %s
    """
      % (self.instance_name, self.header_line, self.header_line,)
    )

    parser.add_argument(
      "input", metavar="CSV(s)", nargs="+", help="The CSV(s) that we wish to reanalyse"
    )
    parser.add_argument(
      "--output",
      metavar="CSV",
      type=str,
      nargs="?",
      default=None,
      help=(
        "Specify output path for length estimation CSV.\n"
        "If more than one input CSV is provided\n"
        "the user will need to provide a comma seperated list\n"
        "with one entry per input. For example:\n\t"
        '--output="zzz.csv, yyy.csv". [Default: None]'
      ),
    )
    parser.add_argument("-s", help=SUPPRESS)
    parser.add_argument("-f", help=SUPPRESS)

    return parser


"""
 These functions are implemented as global
 beacuse multiprocessing does not work with
 class bound methods
"""


def check_results(sim_results):
  if 0 in sim_results:
    print(
      "[WARNING] Telomere length reported zero. This means telomercat\n"
      + "\tfailed to identify enough complete or boundary reads.\n"
      + "\tThis  may mean your original sample was preprocessed to remove \n"
      + "\ttelomere reads. Alternatively this sample could have \n"
      + "\tvery short average TL.",
      file=sys.stderr
    )


def run_simulator(insert_mu, insert_sigma, complete, boundary, proc, read_len, n=10):

  simmer = LengthSimulator(insert_mu, insert_sigma, complete, boundary, read_len)
  res = []
  invalid_count = []
  for i in range(n):
    length, invalid = simmer.start()
    res.append(length)
    invalid_count.append(invalid)

  check_results(res)
  return (np.mean(res), np.std(res))


def estimator_process(job, insert_mu, insert_sigma, complete, boundary, read_len, seed_randomness):

  length_estimator = LengthSimulator(insert_mu, insert_sigma, complete, boundary, read_len, seed_randomness)
  results = length_estimator.start()
  return results


def run_simulator_par(insert_mu, insert_sigma, complete, boundary, proc, read_len, seed_randomness, simulator_n):

  freeze_support()
  #with get_context("forkserver").Pool(proc) as p:
  p = Pool(proc)
  sim_partial = partial(
    estimator_process,
    insert_mu=insert_mu,
    insert_sigma=insert_sigma,
    complete=complete,
    boundary=boundary,
    read_len=read_len,
    seed_randomness=seed_randomness
  )

  results = p.map(sim_partial, range(simulator_n))
  p.close()

  check_results(results)
  return (np.mean(results), np.std(results))


if __name__ == "__main__":
  print ("Do not run this script directly. Type `telomerecat` for help.")
